from typing import Sequence

from opentelemetry.exporter.otlp.proto.grpc import trace_exporter
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceRequest,
)
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.util.instrumentation import (
    InstrumentationInfo,
    InstrumentationScope,
)


class OTLPSpanExporterDD(trace_exporter.OTLPSpanExporter):
    def _translate_data(
        self, data: Sequence[ReadableSpan]
    ) -> ExportTraceServiceRequest:
        for sdk_span in data:
            if "dd_operation_name" in sdk_span.attributes:  # type: ignore[operator]
                # I did not find a good way to get rid of the span kind so these
                # dd_operation_name will almost always get '.internal' appended.
                # I've been told it's not much of a problem.
                sdk_span._instrumentation_scope = InstrumentationScope(
                    name=sdk_span.attributes["dd_operation_name"],  # type: ignore # noqa: E501
                    version="0.0.0",
                )
                sdk_span._instrumentation_info = InstrumentationInfo(
                    name=sdk_span.attributes["dd_operation_name"],  # type: ignore # noqa: E501
                    version="0.0.0",
                )
        return super()._translate_data(data)



from contextlib import contextmanager
from typing import Generator

try:
    from requests import Session, session
except ImportError:
    raise Exception("This feature is only available if 'requests' is installed")

from troncos.traces.propagation import add_context_to_dict


@contextmanager
def traced_session(sess: Session | None = None) -> Generator:
    """
    Creates a requests session that propagates the traceparent forward to the next
    service. Example:

    with traced_session() as session:
        response = session.get("http://postman-echo.com/get")
    """

    s = sess or session()
    add_context_to_dict(s.headers)
    yield s
    del s.headers["traceparent"]



import collections

from opentelemetry import trace
from opentelemetry.trace import Span

from troncos import OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION
from troncos.traces.http import create_http_span, end_http_span
from troncos.traces.propagation import get_propagation_value

SPAN_ATTR_NAME = "_troncos_gunicorn_trace_span"
ACTIVATION_ATTR_NAME = "_troncos_gunicorn_trace_activation"


def _get_span_trace_id(span: Span) -> str:
    context = span.get_span_context()
    if not context:
        return None
    return hex(context.trace_id)[2:]


def pre_request_trace(  # type: ignore[no-untyped-def]
    worker, req, tracer_provider=None
):
    """
    Gunicorn pre request hook function for tracing
    """

    try:
        tp = tracer_provider or trace.get_tracer_provider()
        tracer = tp.get_tracer(OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION)

        peer_ip, peer_port = req.peer_addr

        request_headers = collections.defaultdict(list)
        for k, v in req.headers:
            request_headers[k.lower()].append(v)

        span = create_http_span(
            tracer=tracer,
            http_req_method=req.method,
            http_req_url=req.uri,
            http_req_scheme=req.scheme,
            http_req_flavor=".".join(str(v) for v in req.version),
            http_req_client_ip=peer_ip,
            http_req_client_port=peer_port,
            http_req_headers=request_headers,
            span_name=None,
        )

        # This is usually used as a context manager, but we can't do that here
        # because we only have pre- and post- hooks.
        # So instead we save the activation object and call __enter__ and __exit__
        # manually.
        activation = trace.use_span(span, end_on_exit=True)

        activation.__enter__()

        # What follows is a trick to propagate traces correctly to the rest of the
        # application.
        # There appear to be two kinds of opentelemetry instrumentation:
        #   - client instrumentation (such as requests and psycogp2)
        #   - server instrumentation (such as wsgi and django)
        # Client instrumentation works fine without this hack -- it takes the
        # span parent from the "current span", which we've already set.
        # However, server instrumentation wants to find its trace parent in the
        # _headers_. If there's no incoming traceparent header, this gets
        # really confusing.
        # To avoid that, we create a new traceparent header with this span
        # as the parent, and we _mutate_ the headers that we send onwards to
        # the next level of server instrumentation.
        new_traceparent = get_propagation_value()
        did_set_traceparent = False
        for i, (key, _) in enumerate(req.headers):
            if key.lower() == "traceparent":
                req.headers[i] = (key, new_traceparent)
                did_set_traceparent = True
                break
        if not did_set_traceparent:
            req.headers.append(("TRACEPARENT", new_traceparent))

        setattr(worker, SPAN_ATTR_NAME, span)
        setattr(worker, ACTIVATION_ATTR_NAME, activation)

        return _get_span_trace_id(span)
    except Exception:
        worker.log.exception("Exception generating trace")
        return None


def post_request_trace(worker, req, environ, res):  # type: ignore[no-untyped-def]
    """
    Gunicorn post request hook function for tracing
    """

    try:
        span = getattr(worker, SPAN_ATTR_NAME, None)
        activation = getattr(worker, ACTIVATION_ATTR_NAME, None)

        if not (span and activation):
            worker.log.error(
                "Finishing trace; but worker does not have attached trace info"
            )
            return None

        response_headers = collections.defaultdict(list)
        for k, v in res.headers:
            response_headers[k.lower()].append(v)

        end_http_span(
            span=span,
            http_res_status_code=res.status_code,
            http_res_headers=response_headers,
        )

        activation.__exit__(None, None, None)
        return _get_span_trace_id(span)
    except Exception:
        worker.log.exception("Exception finishing trace")
        return None



from opentelemetry import trace

from troncos import OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION

try:
    from django.utils.module_loading import import_string
except ImportError:
    raise Exception("This feature is only available if 'django' is installed")


def trace_django_middleware(
    middlewares: list[str], tracer_provider: trace.TracerProvider | None = None
) -> None:
    """
    Pass in a list of middlewares to trace
    """

    for m in middlewares:
        klass = import_string(m)
        if klass:
            tp = tracer_provider or trace.get_tracer_provider()
            tr = tp.get_tracer(OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION)
            klass.__call__ = tr.start_as_current_span(m)(klass.__call__)



import opentelemetry.trace as trace

try:
    from structlog.types import EventDict, WrappedLogger
except ImportError:
    raise Exception("This feature is only available if 'structlog' is installed")


def trace_injection_processor(
    logger: WrappedLogger, log_method: str, event_dict: EventDict
) -> EventDict:
    span = trace.get_current_span()
    if not isinstance(span, trace.NonRecordingSpan):
        event_dict["trace_id"] = f"{span.get_span_context().trace_id:x}"
        event_dict["span_id"] = f"{span.get_span_context().span_id:x}"
    return event_dict



from typing import Dict, Literal

from opentelemetry.context import Context
from opentelemetry.propagators import b3, jaeger
from opentelemetry.propagators.textmap import (
    CarrierT,
    DefaultGetter,
    Getter,
    TextMapPropagator,
)
from opentelemetry.trace.propagation import tracecontext


def get_context_from_dict(carrier: CarrierT) -> Context:
    """
    Gets trace context from a dictionary that contains a 'traceparent', 'uber-trace-id'
    or 'b3' entries.
    """

    getter: Getter[CarrierT] = DefaultGetter()  # type: ignore[assignment]
    propagator: TextMapPropagator

    if getter.get(carrier, "traceparent"):  # w3c
        propagator = tracecontext.TraceContextTextMapPropagator()
    elif getter.get(carrier, "uber-trace-id"):  # jaeger
        propagator = jaeger.JaegerPropagator()
    else:  # b3
        # The reason this is default is that multiple headers can be used for
        # propagation.
        propagator = b3.B3MultiFormat()

    return propagator.extract(carrier=carrier)


def add_context_to_dict(
    carrier: CarrierT, fmt: Literal["w3c", "jaeger", "b3"] = "w3c"
) -> CarrierT:
    """
    Adds a trace "parent" entry to a dictionary. This can be in jaeger, b3 or the
    default w3c format.
    """

    propagator: TextMapPropagator
    if fmt == "jaeger":
        propagator = jaeger.JaegerPropagator()
    elif fmt == "b3":
        propagator = b3.B3SingleFormat()
    else:
        propagator = tracecontext.TraceContextTextMapPropagator()

    propagator.inject(carrier)

    return carrier


def get_propagation_value(fmt: Literal["w3c", "jaeger", "b3"] = "w3c") -> str | None:
    """
    Returns the value of the context propagation header
    """

    d: Dict[str, str] = {}
    add_context_to_dict(d, fmt=fmt)

    if fmt == "jaeger":
        return d.get("uber-trace-id")
    elif fmt == "b3":
        return d.get("b3")
    else:
        return d.get("traceparent")



import asyncio
import inspect
import logging
from contextlib import _GeneratorContextManager
from functools import wraps
from types import FunctionType
from typing import Any, Callable, Type, TypeVar, cast

import opentelemetry.trace
from opentelemetry.sdk.resources import Attributes

from troncos import OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION

_TRACE_IGNORE_ATTR = "_trace_ignore"
TFun = TypeVar("TFun", bound=Callable[..., Any])
TClass = TypeVar("TClass")


def _trace_function(
    f: TFun,
    name: str | None = None,
    attributes: Attributes | None = None,
    tracer_provider: opentelemetry.trace.TracerProvider | None = None,
) -> TFun:
    if asyncio.iscoroutinefunction(f):

        @wraps(f)
        async def traced_func_async(*args: tuple, **kwargs: dict[str, Any]) -> Any:
            tp = tracer_provider or opentelemetry.trace.get_tracer_provider()
            tr = tp.get_tracer(OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION)
            with tr.start_as_current_span(
                name or f"{f.__module__}.{f.__qualname__}", attributes=attributes
            ):
                resolved_future = await f(*args, **kwargs)
                return resolved_future

        if hasattr(f, _TRACE_IGNORE_ATTR):
            return f

        return cast(TFun, traced_func_async)
    else:

        @wraps(f)
        def traced_func(*args: tuple, **kwargs: dict[str, Any]) -> Any:
            tp = tracer_provider or opentelemetry.trace.get_tracer_provider()
            tr = tp.get_tracer(OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION)
            with tr.start_as_current_span(
                name or f"{f.__module__}.{f.__qualname__}", attributes=attributes
            ):
                return f(*args, **kwargs)

        if hasattr(f, _TRACE_IGNORE_ATTR):
            return f

        return cast(TFun, traced_func)


def trace_function(
    *args: Any,
    name: str | None = None,
    attributes: Attributes | None = None,
    tracer_provider: opentelemetry.trace.TracerProvider | None = None,
) -> Any:
    """
    This decorator adds tracing to a function. You can supply a tracer provider, if none
    is supplied, the global tracer provider will be used. Example:

    @trace_function
    def myfunc1()
        return "This will be traced"

    @trace_function(tracer_provider=custom_provider)
    def myfunc2()
        return "This will be traced using a custom provider"
    """

    if len(args) > 1:
        raise RuntimeError("Invalid usage of decorator")
    if len(args) == 1 and (callable(args[0]) or asyncio.iscoroutinefunction(args[0])):
        return _trace_function(args[0], name, attributes, tracer_provider)
    else:
        # No args
        def _inner(f: TFun) -> Any:
            return _trace_function(f, name, attributes, tracer_provider)

        return _inner


def trace_block(
    name: str,
    attributes: Attributes | None = None,
    tracer_provider: opentelemetry.trace.TracerProvider | None = None,
) -> _GeneratorContextManager[opentelemetry.trace.Span]:
    """
    Trace using a with statement. You can supply a tracer provider, if none is supplied,
    the global tracer provider will be used. Example:

    with trace_block(name="my block", attributes={"some": "attribute"}):
        time.sleep(1)
    """
    tp = tracer_provider or opentelemetry.trace.get_tracer_provider()
    tr = tp.get_tracer(OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION)
    return tr.start_as_current_span(name, attributes=attributes)


def trace_class(
    *args: Any,
    attributes: Attributes | None = None,
    tracer_provider: opentelemetry.trace.TracerProvider | None = None,
) -> Type[TClass] | Callable[[Type[TClass]], Type[TClass]]:
    """
    This decorator adds a tracing decorator to every method of the decorated class. If
    you don't want some methods to be traced, you can add the 'trace_ignore' decorator
    to them. You can supply a tracer provider, if none is supplied, the global tracer
    provider will be used. Example:

    @trace_class
    class MyClass1:

        def m1(self):
            return "This will be traced"

        @trace_ignore
        def m2(self):
            return "This will not traced"


    @trace_class(tracer_provider=custom_provider)
    class MyClass2:

        def m3(self):
            return "This will be traced using a custom provider"
    """

    def _dec(cls: Type[TClass]) -> Type[TClass]:
        for key, value in cls.__dict__.items():
            if key.startswith("_"):
                continue
            if not (
                isinstance(value, FunctionType) or asyncio.iscoroutinefunction(value)
            ):
                continue

            logging.getLogger(__name__).debug(f"Tracing function {cls.__name__}.{key}")
            setattr(
                cls,
                key,
                _trace_function(
                    value,
                    name=None,
                    attributes=attributes,
                    tracer_provider=tracer_provider,
                ),
            )
        return cls

    if len(args) > 1:
        raise RuntimeError("Invalid usage of decorator")
    if len(args) == 1:
        return _dec(args[0])
    else:
        return _dec


def trace_module(
    attributes: Attributes | None = None,
    tracer_provider: opentelemetry.trace.TracerProvider | None = None,
) -> None:
    """
    This function adds a tracing decorator to every function of the calling module. If
    you don't want some functions to be traced, you can add the 'trace_ignore' decorator
    to them. You can supply a tracer provider, if none is supplied, the global tracer
    provider will be used. Example:

    # Start of module

    def my_function():
        return "This func will be traced"

    @trace_ignore
    def my_function():
        return "This func will not be traced"

    trace_module()

    # End of module
    """

    frame = inspect.stack()[1].frame
    scope = frame.f_locals
    module_name = scope.get("__name__", "unknown")
    for key, value in list(frame.f_locals.items()):
        if key.startswith("_"):
            continue
        if not (isinstance(value, FunctionType) or asyncio.iscoroutinefunction(value)):
            continue
        if getattr(value, "__module__", None) != module_name:
            continue

        logging.getLogger(__name__).debug(f"Tracing function {module_name}.{key}")
        scope[key] = _trace_function(
            value, name=None, attributes=attributes, tracer_provider=tracer_provider
        )


def trace_ignore(f: TFun) -> TFun:
    """
    Decorator to disable automatic tracing of functions.
    See 'trace_module' and 'trace_class'.
    """

    setattr(f, _TRACE_IGNORE_ATTR, ())
    return f



import typing

from opentelemetry.trace import Span, SpanKind, Status, StatusCode, Tracer

from troncos.traces.propagation import get_context_from_dict

SAFE_TRACE_HEADERS = frozenset(
    [
        "accept",
        "cache-control",
        "content-security-policy",
        "content-type",
        "expires",
        "location",
        "origin",
        "range",
        "referer",
        "retry-after",
        "server",
        "traceparent",
        "tracestate",
        "uber-trace-id",
        "x-b3-traceid",
        "x-country",
        "x-language",
        "x-xss-protection",
    ]
)


def create_http_span(
    *,
    tracer: Tracer,
    http_req_method: str,
    http_req_url: str,
    http_req_scheme: str,
    http_req_flavor: str,
    http_req_client_ip: str,
    http_req_client_port: int,
    http_req_headers: typing.Dict[str, list[str]],
    span_name: str | None,
) -> Span:
    """
    Create a new span based on an incoming request. Note that http_req_headers
    key have to be lowercase!
    """

    attr = {
        "http.method": http_req_method,
        "http.url": http_req_url,
        "http.scheme": http_req_scheme,
        "http.flavor": http_req_flavor,
        "net.peer.ip": http_req_client_ip,
        "net.peer.port": http_req_client_port,
    }

    for k, vs in http_req_headers.items():
        if k in SAFE_TRACE_HEADERS:
            normalized_key = k.replace("-", "_")
            attr[f"http.request.header.{normalized_key}"] = (
                vs[0] if len(vs) == 1 else str(vs)
            )

    host_headers = http_req_headers.get("host", ())
    if host_headers:
        attr["http.host"] = host_headers[0]

    user_agent = http_req_headers.get("user-agent", ())
    if user_agent:
        attr["http.user_agent"] = user_agent[0]

    remote_addr = http_req_headers.get("x-forwarded-for", ())
    if remote_addr:
        attr["http.client_ip"] = remote_addr[0]

    return tracer.start_span(
        span_name or f"HTTP {http_req_method}",
        attributes=attr,  # type: ignore[arg-type]
        kind=SpanKind.SERVER,
        context=get_context_from_dict(http_req_headers),
    )


def end_http_span(
    *,
    span: Span,
    http_res_status_code: int,
    http_res_headers: typing.Dict[str, list],
) -> None:
    """
    Add the response information to the traces span. Note that http_res_headers
    key have to be lowercase!
    """

    if res_length := http_res_headers.get("content-length"):
        span.set_attribute("http.response_content_length", res_length[0])

    for k, vs in http_res_headers.items():
        if k in SAFE_TRACE_HEADERS:
            normalized_key = k.replace("-", "_")
            span.set_attribute(
                f"http.response.header.{normalized_key}",
                vs[0] if len(vs) == 1 else str(vs),
            )

    span.set_attribute("http.status_code", http_res_status_code)

    if 200 <= http_res_status_code <= 399:
        span.set_status(
            Status(
                status_code=StatusCode.OK,
            )
        )
    else:
        span.set_status(
            Status(
                status_code=StatusCode.ERROR,
                description=str(http_res_status_code),
            )
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['troncos',
 'troncos.frameworks',
 'troncos.frameworks.django',
 'troncos.frameworks.gunicorn',
 'troncos.frameworks.requests',
 'troncos.frameworks.starlette',
 'troncos.frameworks.structlog',
 'troncos.logs',
 'troncos.traces']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.0,<2.0',
 'opentelemetry-exporter-otlp-proto-grpc>=1.12.0,<2.0.0',
 'opentelemetry-propagator-b3>=1.12.0,<2.0.0',
 'opentelemetry-propagator-jaeger>=1.12.0,<2.0.0']

setup_kwargs = {
    'name': 'troncos',
    'version': '0.0.17',
    'description': 'Observability tools and boilerplate for use in python apps',
    'long_description': '<h1 align="center" style="border-bottom: 0">\n  🪵<br>\n  Troncos <br/>\n</h1>\n\n<p align="center">\n    <em>\n        Collection of Python logging and tracing tools\n    </em>\n    <br>\n    <a href="https://github.com/kolonialno/troncos/actions?workflow=CI">\n        <img src="https://github.com/kolonialno/troncos/actions/workflows/ci.yml/badge.svg" alt="CI status">\n    </a>\n    <a href="https://pypi.python.org/pypi/troncos">\n        <img src="https://img.shields.io/pypi/v/troncos.svg">\n    </a>\n    <img src="https://img.shields.io/pypi/pyversions/troncos">\n    <a href="https://github.com/kolonialno/troncos/blob/master/LICENSE">\n        <img src="https://img.shields.io/github/license/kolonialno/troncos.svg">\n    </a>\n</p>\n\n<!-- TOC -->\n  * [Installation](#installation)\n  * [Etymology](#etymology)\n  * [Setup](#setup)\n    * [Plain](#plain)\n    * [Starlette (with uvicorn)](#starlette--with-uvicorn-)\n    * [Django (with gunicorn)](#django--with-gunicorn-)\n  * [Logging](#logging)\n    * [Structlog](#structlog)\n  * [Tracing](#tracing)\n    * [trace_function](#trace_function)\n    * [trace_block](#trace_block)\n    * [trace_class](#trace_class)\n    * [trace_module](#trace_module)\n    * [trace_ignore](#trace_ignore)\n    * [Other instrumentors for tracing](#other-instrumentors-for-tracing)\n  * [Trace Propagation](#trace-propagation)\n    * [Send context](#send-context)\n      * [Requests](#requests)\n      * [Manually](#manually)\n    * [Receive context](#receive-context)\n      * [Using troncos middleware](#using-troncos-middleware)\n      * [Manually](#manually)\n<!-- TOC -->\n\n## Installation\n\n```console\n# With pip\n$ pip install troncos\n\n# With poetry\n$ poetry add troncos\n```\n\n## Etymology\n\n"Troncos" is the plural of the spanish word "Tronco", which translates to "trunk" or "log".\n\n## Setup\n\n> **NOTE**: It is a good idea to use a `settings.py`-file (or similar) as an authoritative source of variables (service name, environment, whether tracing is enabled or not, log level etc.)\n\n### Plain\n\nSetting up logging and tracing requires some code that lives as close to the invocation of the application as possible (like in an entrypoint).\n\n```python\nfrom os import environ\n\nfrom troncos.logs import init_logging_basic\nfrom troncos.traces import init_tracing_basic\n\ninit_logging_basic(\n    level=environ.get("LOG_LEVEL", "INFO"),\n    formatter=environ.get("LOG_FORMATTER", "cli")  # Use "logfmt" or "json" in k8s\n)\n\ninit_tracing_basic(\n    endpoint=environ.get("TRACING_PATH", "http://localhost:4317"),\n    attributes={\n        "environment": environ.get("ENVIRONMENT", "localdev"),\n        "service.name": "myservice",\n    }\n)\n\n# Add other instrumentors here, like:\n# RequestsInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n#     "service.name": "requests",\n# }))\n```\n\n### Starlette (with uvicorn)\n\n```python\nfrom os import environ\n\nfrom troncos.frameworks.starlette.uvicorn import init_uvicorn_observability\nfrom troncos.logs import init_logging_basic\nfrom troncos.traces import init_tracing_basic\n\ninit_logging_basic(\n    level=environ.get("LOG_LEVEL", "INFO"),\n    formatter=environ.get("LOG_FORMATTER", "cli")  # Use "logfmt" or "json" in k8s\n)\n\ninit_tracing_basic(\n    endpoint=environ.get("TRACING_PATH", "http://localhost:4317"),\n    attributes={\n        "environment": environ.get("ENVIRONMENT", "localdev"),\n        "service.name": "myservice",\n    }\n)\n\n# Add other instrumentors here, like:\n# RequestsInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n#     "service.name": "requests",\n# }))\n\napp = ...  # Setup your app\n\ninit_uvicorn_observability(\n    app=app,\n    log_access_ignored_paths=["/health", "/metrics"],  # Do not log these requests\n)\n```\n\n### Django (with gunicorn)\n\nTo set up tracing you have to set up some gunicorn hooks. Create a `gunicorn/config.py` file in your project:\n\n```python\nfrom os import environ\n\nfrom troncos.frameworks.gunicorn import post_request_trace, pre_request_trace\nfrom troncos.traces import init_tracing_basic \n\n\ndef post_fork(server, worker):\n    init_tracing_basic(\n      endpoint=environ.get("TRACING_PATH", "http://localhost:4317"),\n      attributes={\n        "pid": worker.pid,\n        "environment": environ.get("ENVIRONMENT", "localdev"),\n        "service.name": "myservice",\n      }\n    )\n\n    # Add other instrumentors here, like:\n    # DjangoInstrumentor().instrument()\n    #\n    # Psycopg2Instrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    #     "service.name": "psycopg2",\n    # }))\n\ndef pre_request(worker, req):\n    pre_request_trace(worker, req)\n\n\ndef post_request(worker, req, environ, resp):\n    post_request_trace(worker, req, environ, resp)\n```\n\nThen when running gunicorn, specify what config file to use:\n\n```console\ngunicorn myapp.wsgi:application --config python:myapp.gunicorn.config ...\n```\n\nYou have to manually configure logging in your `settings.py`. You should adhere to the principle described in [the logging section](#logging).\n\nMake sure that you add the `TraceIdFilter` to all handlers. Your logging configuration should look roughly like this:\n\n```python\nfrom os import environ\n\nLOGGING = {\n    "version": 1,\n    "disable_existing_loggers": True,\n    "filters": {\n        "trace_id": {"()": "troncos.logs.filters.TraceIdFilter"},\n    },\n    "formatters": {\n        "cli": {"()": "troncos.logs.formatters.PrettyFormatter"},\n        "json": {"()": "troncos.logs.formatters.JsonFormatter"},\n        "logfmt": {"()": "troncos.logs.formatters.LogfmtFormatter"},\n    },\n    "handlers": {\n        "console": {\n            "class": "logging.StreamHandler",\n            "formatter": environ.get("LOG_FORMATTER", "logfmt"),\n            "filters": ["trace_id"],\n        }\n    },\n    "loggers": {\n        "interno": {"handlers": ["console"], "level": environ.get("LOG_LEVEL", "INFO")},\n        "django": {"handlers": ["console"], "level": environ.get("LOG_LEVEL", "INFO")},\n        "django.server": {\n            "handlers": ["console"],\n            "level": environ.get("LOG_LEVEL", "INFO"),\n            "propagate": False,\n        },\n        "gunicorn.error": {\n            "handlers": ["console"], \n            "level": environ.get("LOG_LEVEL", "INFO")\n        },\n    },\n}\n```\n\n## Logging\n\nMore often then not, you want all loggers to propagate their records to the `root` logger and make the `root` logger handle everything. Depending on your project, this might require some additional configuration. Looking at the [python logging flow](https://docs.python.org/3/howto/logging.html#logging-flow) can help you understand how child loggers can propagate records to the `root` logger. Note that propagating to `root` is the default behaviour.\n\nThere is a nice helper function that will print all loggers in troncos called `print_loggers`:\n\n```python\nfrom troncos.logs import print_loggers\nprint_loggers(verbose=False)  # To visualize loggers\n```\n\nAfter calling the `init_logging_basic` function in a simple project you should see something like this printed `print_loggers`:\n\n```console\nLoggers:\n[ root                 ] logging.RootLogger LEVEL: 20 PROPAGATE: True\n  └ HANDLER logging.StreamHandler  LEVEL: 20\n    └ FILTER troncos.logs.filters.TraceIdFilter\n    └ FORMATTER troncos.logs.formatters.PrettyFormatter\n```\n\nSo in general, after the initial setup you can use any logger and that will propagate the log record to root:\n\n```python\nimport logging\n\nlogging.getLogger("my.random.logger").info("Root will handle this record")\n```\n\n### Structlog\n\nTo include traces in your structlog logs, add this processor to your configuration.\n\n> **NOTE**: This only adds trace information to your logs if you have set up tracing in your project.\n\n```python\nimport structlog\n\nfrom troncos.frameworks.structlog.processors import trace_injection_processor\n\nstructlog.configure(\n    processors=[\n       trace_injection_processor,\n    ],\n)\n```\n\n## Tracing\n\nAfter initializing tracing in your project you can use different methods to trace your code.\n\n### trace_function\n\nThis decorator adds tracing to a function. You can supply a tracer provider, if none is supplied, the global tracer provider will be used:\n\n```python\nfrom troncos.traces.decorate import trace_function\nfrom troncos.traces import init_tracing_provider\n\ncustom_provider = init_tracing_provider(attributes={\n  "service.name": "my_custom_provider",\n})\n\n@trace_function\ndef myfunc1():\n    return "This will be traced"\n\n@trace_function(tracer_provider=custom_provider)\ndef myfunc2():\n    return "This will be traced using a custom provider"\n```\n\n### trace_block\n\nTrace using a with statement. You can supply a tracer provider, if none is supplied, the global tracer provider will be used.\n\n```python\nfrom troncos.traces.decorate import trace_block\n\nwith trace_block(name="my block", attributes={"some": "attribute"}):\n    print("... do something ...")\n```\n\n### trace_class\n\nThis decorator adds a tracing decorator to every method of the decorated class. If you don\'t want some methods to be traced, you can add the [trace_ignore](#trace_ignore) decorator to them. You can supply a tracer provider, if none is supplied, the global tracer provider will be used:\n\n```python\nfrom troncos.traces.decorate import trace_class, trace_ignore\nfrom troncos.traces import init_tracing_provider\n\ncustom_provider = init_tracing_provider(attributes={\n    "service.name": "my_custom_provider",\n})\n\n@trace_class\nclass MyClass1:\n\n    def m1(self):\n        return "This will be traced"\n\n    @trace_ignore\n    def m2(self):\n        return "This will not traced"\n\n\n@trace_class(tracer_provider=custom_provider)\nclass MyClass2:\n\n    def m3(self):\n        return "This will be traced using a custom provider"\n```\n\n### trace_module\n\nThis function adds a tracing decorator to every function of the calling module. If you don\'t want some functions to be traced, you can add the [trace_ignore](#trace_ignore) decorator to them. You can supply a tracer provider, if none is supplied, the global tracer provider will be used:\n\n```python\nfrom troncos.traces.decorate import trace_ignore, trace_module\n\ndef my_function():\n    return "This func will be traced"\n\n@trace_ignore\ndef my_function():\n    return "This func will not be traced"\n\ntrace_module()\n```\n\n### trace_ignore\n\nA decorator that will make [trace_class](#trace_class) and [trace_module](#trace_module) ignore the decorated function/method.\n\n### Other instrumentors for tracing\n\nYou can add extra instrumentors to you app for even more tracing. You have to install the relevant packages yourself.\n\n```python\nfrom troncos.traces import init_tracing_provider\n\nDjangoInstrumentor().instrument()\n\nPsycopg2Instrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "psycopg2",\n}))\n\nRedisInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "redis",\n}))\n\nCeleryInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "celery",\n}))\n\nElasticsearchInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "elasticsearch",\n}))\n\nGrpcInstrumentorClient().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "grpc",\n}))\n\nRequestsInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "requests",\n}))\n\nHTTPXClientInstrumentor().instrument(tracer_provider=init_tracing_provider(attributes={\n    "service.name": "requests",  # Async requests\n}))\n```\n\n## Trace Propagation\n\nIf you want to propagate your trace to the next service, you need to send/receive the `traceparent` header with your requests/message. Here are examples on how to do that.\n\n### Send context\n\n#### Requests\n\nIn general, if you have the `RequestsInstrumentor` setup you do not have to think about this. If you are not using that for some reason, you can propagate with this method:\n\n```python\nimport requests\nfrom troncos.frameworks.requests import traced_session\n\n# Using a new session\nwith traced_session() as s:\n    response = s.get("http://postman-echo.com/get")\n\n\n# Using an old session\nmy_session = requests.session()\nwith traced_session(my_session) as s:\n    response = s.get("http://postman-echo.com/get")\n```\n\n#### Manually\n\n```python\nfrom troncos.traces.propagation import get_propagation_value\n\n# Get traceparent\ntraceparent = get_propagation_value()\n\n# Send it somewhere\n```\n\nor\n\n```python\nfrom troncos.traces.propagation import add_context_to_dict\n\nsome_dict = {}\n\n# Add traceparent to dict\nadd_context_to_dict(some_dict)\n\n# Send it somewhere\n```\n\n### Receive context\n\n#### Using troncos middleware\n\nTroncos defines middleware for some frameworks that does this automatically for you. If your framework is missing in troncos, please create an issue or PR.\n\n#### Manually\n\n```python\nfrom troncos.traces.propagation import get_context_from_dict\nfrom opentelemetry.trace import get_tracer\n\nsome_dict = ...\ncontext = get_context_from_dict(some_dict)\n\nwith get_tracer(__name__).start_as_current_span(\n        "span.name",\n        attributes={"some": "attrs"},\n        context=context,\n):\n    print("... do something ...")\n```',
    'author': 'Guðmundur Björn Birkisson',
    'author_email': 'gudmundur.birkisson@oda.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kolonialno/troncos',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import logging

import opentelemetry.trace as trace


class TraceIdFilter(logging.Filter):
    """
    Simple logging filter that adds a trace_id to the log record if available.

    It sounds strange to use filters to modify records, but this is described
    in the official python docs:

    https://docs.python.org/3/howto/logging-cookbook.html#using-filters-to-impart-contextual-information
    """

    def __init__(self, name: str = "TraceIdFilter") -> None:
        super().__init__(name)

    def filter(self, record: logging.LogRecord) -> bool:
        span = trace.get_current_span()
        if not isinstance(span, trace.NonRecordingSpan):
            record.trace_id = f"{span.get_span_context().trace_id:x}"
            record.span_id = f"{span.get_span_context().span_id:x}"
        return True


class HttpPathFilter(logging.Filter):
    """
    Simple logging filter that drops any log records regarding paths in the
    list 'ignored_paths'.
    """

    def __init__(
        self, ignored_paths: list[str] | None, name: str = "HttpPathFilter"
    ) -> None:
        super().__init__(name)
        self._ignored_paths = ignored_paths if ignored_paths else []

    def filter(self, record: logging.LogRecord) -> bool:
        if hasattr(record, "http_path"):
            return getattr(record, "http_path") not in self._ignored_paths
        return True



import collections
import logging

from opentelemetry.trace import TracerProvider, use_span

from troncos import OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION
from troncos.traces.http import create_http_span, end_http_span

try:
    from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.types import ASGIApp
except ImportError:
    raise Exception("This feature is only available if 'starlette' is installed")


class TracingMiddleware(BaseHTTPMiddleware):
    """
    Tracing middleware for starlette
    """

    def __init__(
        self,
        app: ASGIApp,
        tracer_provider: TracerProvider,
        span_name: str | None = None,
    ) -> None:
        super().__init__(app)
        self._span_name = span_name
        self._tracer = tracer_provider.get_tracer(
            OTEL_LIBRARY_NAME, OTEL_LIBRARY_VERSION
        )

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        client_ip, client_port = request.client or ("NOIP", -1)

        request_headers = collections.defaultdict(list)
        for k, v in request.headers.items():
            request_headers[k.lower()].append(v)

        span = create_http_span(
            tracer=self._tracer,
            http_req_method=request.method,
            http_req_url=str(request.url),
            http_req_scheme=request.url.scheme,
            http_req_flavor="".join(
                str(v) for v in request.scope.get("http_version", [])
            ),
            http_req_client_ip=client_ip,
            http_req_client_port=client_port,
            http_req_headers=request_headers,
            span_name=self._span_name,
        )
        with use_span(span, end_on_exit=True):
            response: Response = await call_next(request)

            response_headers = collections.defaultdict(list)
            for k, v in response.headers.items():
                response_headers[k.lower()].append(v)

            end_http_span(
                span=span,
                http_res_status_code=response.status_code,
                http_res_headers=response_headers,
            )

            return response


class LoggingMiddleware(BaseHTTPMiddleware):
    """
    This is our custom logs middleware for starlette. We need this so trace_id is
    added to our log records.
    """

    def __init__(
        self,
        app: ASGIApp,
        access_logger: logging.Logger,
        error_logger: logging.Logger,
    ) -> None:
        super().__init__(app)
        self._access = access_logger
        self._error = error_logger

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        req_cli_ip, req_cli_port = request.scope.get("client")
        try:
            response: Response = await call_next(request)

            self._access.info(
                "",
                extra={
                    "http_client_addr": f"{req_cli_ip}:{req_cli_port}",
                    "http_method": request.scope.get("method"),
                    "http_path": request.scope.get("path"),
                    "http_version": request.scope.get("http_version"),
                    "http_status_code": response.status_code,
                },
            )
        except Exception as e:
            self._error.error(
                "",
                exc_info=e,
                extra={
                    "http_client_addr": f"{req_cli_ip}:{req_cli_port}",
                    "http_method": request.scope.get("method"),
                    "http_path": request.scope.get("path"),
                    "http_version": request.scope.get("http_version"),
                    "http_status_code": 500,
                },
            )
            raise e
        return response



import logging
from typing import Iterable, List

import opentelemetry.trace
from opentelemetry.exporter.otlp.proto.grpc import trace_exporter
from opentelemetry.sdk.resources import Attributes, Resource
from opentelemetry.sdk.trace import SpanProcessor, TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
    ConsoleSpanExporter,
    SimpleSpanProcessor,
)
from opentelemetry.util._once import Once

from troncos.traces.dd_exporter import OTLPSpanExporterDD

_GLOBAL_SPAN_PROCESSORS: list[SpanProcessor] | None = None
_GLOBAL_SPAN_PROCESSORS_SET_ONCE = Once()

_DEBUG_SPAN_PROCESSOR: SpanProcessor = SimpleSpanProcessor(ConsoleSpanExporter())


def _set_span_processors(span_processors: list[SpanProcessor]) -> None:
    def set_sp() -> None:
        global _GLOBAL_SPAN_PROCESSORS
        _GLOBAL_SPAN_PROCESSORS = span_processors

    did_set = _GLOBAL_SPAN_PROCESSORS_SET_ONCE.do_once(set_sp)
    if not did_set:
        logging.getLogger(__name__).warning(
            "Global span processors already set, not doing that again!"
        )


def init_tracing_endpoints(
    endpoint: str | None, endpoint_dd: str | None = None
) -> list[SpanProcessor]:
    """
    Initialize the global span processor.
    """
    exporters = []

    if endpoint:
        otel_exp = trace_exporter.OTLPSpanExporter(endpoint=endpoint)
        logging.getLogger(__name__).info(
            "Reporting OTEL traces with %s(endpoint=%s)",
            type(otel_exp).__name__,
            endpoint,
        )
        exporters.append(BatchSpanProcessor(otel_exp))

    if endpoint_dd:
        dd_exp = OTLPSpanExporterDD(endpoint=endpoint_dd)
        logging.getLogger(__name__).info(
            "Reporting DD traces with %s(endpoint=%s)",
            type(dd_exp).__name__,
            endpoint_dd,
        )
        exporters.append(BatchSpanProcessor(dd_exp))

    _set_span_processors(exporters)  # type: ignore[arg-type]
    return _GLOBAL_SPAN_PROCESSORS  # type: ignore[return-value]


def init_tracing_provider(
    attributes: Attributes, global_provider: bool = False
) -> TracerProvider:
    """
    Initialize a tracing provider. By default, this function will make the new tracer
    provider the global one. If that is not desired, pass in global_provider=False.
    """

    if not _GLOBAL_SPAN_PROCESSORS:
        raise RuntimeError("Call 'init_tracing_endpoint' before calling this function")

    if not attributes.get("service.name"):
        raise ValueError("Tracer must have 'service.name' in attributes")

    if global_provider and not attributes.get("environment"):
        raise ValueError("Global tracer must have 'environment' in attributes")

    resource = Resource.create(attributes)
    provider = TracerProvider(resource=resource)
    for span_processor in _GLOBAL_SPAN_PROCESSORS:
        provider.add_span_processor(span_processor)

    if global_provider:
        opentelemetry.trace.set_tracer_provider(provider)

    return provider


def init_tracing_debug(
    trace_provider: TracerProvider | List[TracerProvider],
) -> None:
    """
    Add debug processor to tracing providers.
    """

    if isinstance(trace_provider, Iterable):
        for p in trace_provider:
            p.add_span_processor(_DEBUG_SPAN_PROCESSOR)
    else:
        trace_provider.add_span_processor(_DEBUG_SPAN_PROCESSOR)


def init_tracing_basic(
    endpoint: str | None = None,
    endpoint_dd: str | None = None,
    attributes: Attributes | None = None,
    debug: bool = False,
) -> TracerProvider:
    """
    Setup rudimentary tracing.
    """

    init_tracing_endpoints(endpoint, endpoint_dd)
    global_tracer = init_tracing_provider(attributes or {}, global_provider=True)
    if debug:
        init_tracing_debug(global_tracer)
    return global_tracer
